///////////////////////////////////////////////////////////////////////////
// 
// File						: Audio.cpp
//
// Description				: Audio functions implementation
//
// Author					: Tomasz Stachowiak
//
// Date						: 11-15-96
//
////////////////////////////////////////////////////////////////////////////
#include <stdio.h>

#include "stdafx.h"
#include "global.h"
#include "gsm.h"
#include "BVA.h"
#include "UserData.h"
#include "rtp.h"
#include "TSaudioProc.h"
#include "BuenaVista.h"
#include "BuenaVistaDlg.h"
#include "Audio.h"
#include "AudioStat.h"
#include "options.h"
#include "conference.h"

#define IDM_EXIT                        32771

CBuenaVistaDlg *global_dlg;
Audio *cf;

////////////////////////////////////////////////////////////////////////////////
//
// Exitting callback
//
////////////////////////////////////////////////////////////////////////////////
void Exit()
{
	SendMessage(global_dlg->m_hWnd, WM_CLOSE, 0, NULL); 
}

////////////////////////////////////////////////////////////////////////////////
//
// Audio constructor
//
////////////////////////////////////////////////////////////////////////////////
Audio::Audio(CBuenaVistaDlg *dlg, Session *session)
{
	int i;
	HANDLE hProcess;

	main_dlg = dlg;
	global_dlg = dlg;
 	options = &(((Conference *)main_dlg->conference)->options.audio);
 
	hProcess = GetCurrentProcess();
	//SetPriorityClass(hProcess, HIGH_PRIORITY_CLASS);
	SetPriorityClass(hProcess, REALTIME_PRIORITY_CLASS);
	
	code_type = GSM;
	mode = DUPLEX;
	treshold = 2000;
	timer = 0;
	silence_no = 0;
	silence_mic = 0;
	signal = 0;
	current_sender = -1;
	talk = 1;
	manual = 0;

	// ADPCM init
	adpcmsendstate.valprev = 0;
	adpcmsendstate.index = 0;

	for(i=0; i<MAX_USERS; i++)
	{
		adpcmstates[i].valprev = 0;
		adpcmstates[i].index = 0;
		first[i] = 1;
		bvabuffer[i] = NULL;
		userdata[i] = NULL;
    }

	gsmHandle = gsm_create();
	
	// Start up NCS
	ncsApp = (ncsApplication *)ncsAppAttach(AUDIO_PLUGIN_PORT, session);
	
	if(ncsApp == NULL)
	{
		Exit();
		return;
	}

	if(!hardwareInit())
		SendMessage(main_dlg->m_hWnd, WM_CLOSE, 0, NULL); 


	if(mode == HALF_DUPLEX)
	{
		automatic = 0;
//		main_dlg->m_Auto.SetCheck(0);
		state = stateMic;
	}
	else
	{
		automatic = 0;
//		main_dlg->m_Auto.SetCheck(0);
		state = stateBoth;
		main_dlg->EnableAudioCheck(FALSE);
		main_dlg->SetAudioCheck(TRUE);
		//setMic();
	}

	// Create reading thread
	hThread = CreateThread( 
		    NULL,									// no security attributes 
			0,										// use default stack size  
			(LPTHREAD_START_ROUTINE) ReadThread,	// thread function 
			this,									// argument to thread function 
			0,										// use default creation flags 
			&dwThreadId);							// returns the thread identifier 
	//SetThreadPriority(hThread, THREAD_PRIORITY_TIME_CRITICAL);
	SetThreadPriority(hThread, THREAD_PRIORITY_HIGHEST);
	
	if (hThread == NULL) 
	{
		MessageBox(NULL, "CreateThread error", "Error", MB_OK);
	}

	cf = this;

}

////////////////////////////////////////////////////////////////////////////////
//
// Audio destructor
//
////////////////////////////////////////////////////////////////////////////////
Audio::~Audio(void)
{
	FILE *file;
	int i;
	
	TerminateThread(hThread, 0);
	CloseHandle(hThread);

	switch(state)
	{
	case stateMic:
		audioReadClose();
		break;

	case stateSpeaker:
		bva->close();
		ncsAppSendData(ncsApp, -1, STOP, NULL, 0);
		break;

	case stateBoth:
		audioReadClose();
		bva->close();
		ncsAppSendData(ncsApp, -1, STOP, NULL, 0);
		break;
	}

	delete bva;
	gsm_destroy(gsmHandle);
	audioClose();
	ncsAppDettach(ncsApp);

	file = fopen("userdata.log", "w");
	fprintf(file, "AUDIO RECEIVE REPORT\n");
	for(i=0; i<MAX_USERS; i++)
		if(userdata[i] != NULL)
			fprintf(file, "User %d, packets: received %d, lost %d, late %d\n", i, userdata[i]->frames_received, userdata[i]->frames_lost, userdata[i]->frames_late);
	fclose(file);

}

////////////////////////////////////////////////////////////////////////////////
//
// Audio initializing procedure
//
////////////////////////////////////////////////////////////////////////////////
int Audio::hardwareInit()
{
	// audio init
	audioInit(this);
	bva = new BVA();
	if(!bva->init())
	{
		main_dlg->MessageBox("There is no appropriate audio device", "Info", MB_OK | MB_ICONEXCLAMATION);
		return 0;
	}

	bva->close();

	if(!audioReadInit())
	{
		main_dlg->MessageBox("Audio device does not support audio capturing", "Info", MB_OK | MB_ICONEXCLAMATION);
		bva->init();
	}
	else
		if(!bva->init())
		{
			mode = HALF_DUPLEX;
			//main_dlg->MessageBox("Audio device supports only HALF DUPLEX mode", "Info", MB_OK | MB_ICONEXCLAMATION);
		}
	else
		main_dlg->EnableAudioCheck(FALSE);

	return 1;
}

////////////////////////////////////////////////////////////////////////////////
//
// Main conference reading thread
//
////////////////////////////////////////////////////////////////////////////////
DWORD Audio::ReadThread(LPDWORD lpdwParam)
{
	struct timeval timeout;
	fd_set set;
	RTPheader header;
	char *msg;
	int size;
	int sockfd;
	char *data;
	ncsApplication *app = ((Audio *)lpdwParam)->ncsApp;

	timeout.tv_sec = 0;
	timeout.tv_usec = 1000;

	while(1)
	{
		sockfd = ncsAppGetFD(app, &set);
		if(select(sockfd+1, &set, 0, 0, &timeout) > 0)
		{
			ncsAppProcessIncoming(app, &set);
			size = ncsAppGetData(app, &set, &data);
			if(size > 0)
			{
				size = size - RTP_HEADER_SIZE;
				readRTPheader(&header, data);
				msg = data + RTP_HEADER_SIZE;
				((Audio *)lpdwParam)->HandleData(&header, msg, size);
				free(data);
			}
		}
	}
	return 1;
}

////////////////////////////////////////////////////////////////////////////////
//
// Audio receiving procedure
//
////////////////////////////////////////////////////////////////////////////////
void Audio::HandleData(RTPheader *header, void *data, int len)
{
	char buffer[2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE];
	short level;
	int i;
	DWORD handle, tag;

	tag = header->pt;
	handle = header->ssrc;
	
	if(userdata[handle] == NULL)
		userdata[handle] = new UserData(header, len);
	else
		if(!userdata[handle]->frame(header, len))
			return;

	((CAudioStat *)main_dlg->audioStat)->update(handle, userdata[handle]);

	switch (tag)
	{
	case ADPCM:
		memcpy(&adpcmstates[handle].valprev, data, 2);
		adpcmstates[handle].valprev = ntohs(adpcmstates[handle].valprev);
		adpcmstates[handle].index = ((char *)data)[2];

		data = (char *)data + 3;

		adpcm_decoder((char *)data, (short *)buffer, AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE, &adpcmstates[handle]);
		break;

    case GSM:
		for(i=0; i<AUDIO_CHUNKS_NO; i++)
		{
			gsm_decode(gsmHandle, ((gsm_frame *)data)[i], (short *)&buffer[i * 320]);
		}

		break;

	case PCM:
		memcpy(buffer, data, 2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
		for(i=0; i<AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE; i++)
			((short*)buffer)[i] = ntohs(((short *)buffer)[i]);
		break;

	case STOP:
		if(userdata[handle] != NULL)
		{
			delete userdata[handle];
			userdata[handle] = NULL;
		}

		if(bvabuffer[handle] != NULL)
		{
			bva->deleteBuffer(bvabuffer[handle]);
			bvabuffer[handle] = NULL;
			return;
		}
		break;

	}
	
	if(options->mute)
		ZeroMemory(buffer, 2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

	level = audioGetLevel(buffer);
	main_dlg->SetSpeakerInd(level/330);
	
	switch(state)
	{
		case stateMic:
			if(level > treshold)
			{
				signal = 1;
				//stateControl(sig);
				//current_sender = handle;
				//audioWriteChunk(buffer);
			}
			break;

		case stateSpeaker:
			if(level > treshold)
				checkTimer();
			
			if(bvabuffer[handle] == NULL)
				bvabuffer[handle] = bva->createBuffer(100*2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
			if(bvabuffer[handle] == NULL)
			{
				MessageBox(NULL, "BV audio buffer init failed!", "Error", MB_OK);
				return;
			}
			bvabuffer[handle]->play(buffer, 2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);

			break;

		case stateBoth:

			if(bvabuffer[handle] == NULL)
				bvabuffer[handle] = bva->createBuffer(100*2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
			if(bvabuffer[handle] == NULL)
			{
				MessageBox(NULL, "BV audio buffer init failed!", "Error", MB_OK);
				return;
			}
			bvabuffer[handle]->play(buffer, 2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
			break;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Audio sending procedure
//
////////////////////////////////////////////////////////////////////////////////
void Audio::SendData(char *data)
{
	char *msg;
	gsm_frame *gsm_msg;
	short *buf;
	int i;
	short level;
	short tmp;

	if(options->micOff)
		ZeroMemory(data, BASIC_AUDIO_SIZE*AUDIO_CHUNKS_NO*2);

	level = audioGetLevel(data);
	if(level >= mic_treshold)
	{
		switch(options->compression)
		{
		case GSM:
			gsm_msg = (gsm_frame *)malloc(AUDIO_CHUNKS_NO * sizeof(gsm_frame));
			for(i=0; i<AUDIO_CHUNKS_NO; i++)
			{
				gsm_encode(gsmHandle, (gsm_signal *)(&data[320 * i]), gsm_msg[i]);
			}
			ncsAppSendData(ncsApp, -1, GSM, gsm_msg, AUDIO_CHUNKS_NO*sizeof(gsm_frame));
			free(gsm_msg);
			break;
	      
		case ADPCM:
			msg = (char *)malloc(BASIC_AUDIO_SIZE*AUDIO_CHUNKS_NO/2 + 3);
			if (!msg)
			{
				fprintf(stderr, "Message allocation error\n");
				return;
			}
			tmp = htons(adpcmsendstate.valprev);
			memcpy(msg, &tmp, 2);
			msg[2] = adpcmsendstate.index;
			msg += 3;
			adpcm_coder((short *)data, msg, BASIC_AUDIO_SIZE*AUDIO_CHUNKS_NO, &adpcmsendstate);
			msg -= 3;
			ncsAppSendData(ncsApp, -1, ADPCM, msg, BASIC_AUDIO_SIZE*AUDIO_CHUNKS_NO/2 + 3);
			free(msg);
			break;
	
		case PCM:
			buf = (short *)malloc(2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
			memcpy(buf, data, 2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
			for(i=0; i<AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE; i++)
				buf[i] = htons(buf[i]);
			ncsAppSendData(ncsApp, -1, PCM, buf, 2*AUDIO_CHUNKS_NO*BASIC_AUDIO_SIZE);
			free(buf);
			break;
	
		}		
   	}
	//else
	//	ncsAppSendData(ncsApp, -1, STOP, NULL, 0);


	main_dlg->SetMicInd(level/330);
	
	if(level <= mic_treshold)
	{
		silence_mic++;
		if(silence_mic >= 10)
		{
			checkTimer();
			if(signal)
			{
				silence_mic = 0;
				stateControl(sig);
			}
		}
	}
	else
		silence_mic = 0;

}

////////////////////////////////////////////////////////////////////////////////
//
// GUI callbacks
//
////////////////////////////////////////////////////////////////////////////////
void Audio::setTreshold(UINT level)
{
	treshold = level * 327;
}

void Audio::setMicTreshold(int level)
{
	mic_treshold = level * 327;
}

void Audio::codeADPCM(void)
{
	int i;
	
	for(i=0; i<MAX_USERS; i++)
	{
		adpcmstates[i].valprev = 0;
		adpcmstates[i].index = 0;
    }
	code_type = ADPCM;
}

void Audio::codeGSM(void)
{
	code_type = GSM;
}

void Audio::codePCM()
{
	code_type = PCM;
}

void Audio::setAuto(int check)
{
	if(check != automatic)
		stateControl(Abut);
}

void Audio::setManual(int check)
{
	if(check != manual)
		stateControl(Mbut);
}

void Audio::setTalk(int check)
{
	talk = check;
	stateControl(Sbut);
}

////////////////////////////////////////////////////////////////////////////////
//
// Signal stoping timer procedures
//
////////////////////////////////////////////////////////////////////////////////
void Audio::checkTimer()
{
	if(timer != 0)
		KillTimer(global_dlg->m_hWnd, 23000);

	timer = SetTimer(global_dlg->m_hWnd, 23000, 1000, TimerCallback);
	if(timer == 0)
		MessageBox(NULL, "Timer creating error!", "Error", MB_OK);
}

VOID CALLBACK Audio::TimerCallback(
    HWND hwnd,	// handle of window for timer messages 
    UINT uMsg,	// WM_TIMER message
    UINT idEvent,	// timer identifier
    DWORD dwTime 	// current system time
   )
{
	cf->stateControl(Nsig);
	cf->signal = 0;
	cf->silence_no = 0;
}


///////////////////////////////////////////////////////////////////////////
// 
// Automaton state transition
//
///////////////////////////////////////////////////////////////////////////
void Audio::stateControl(StateInput in)
{
	if(mode == DUPLEX)
	{
		FDstateControl(in);	// full-duplex state control
	}

	if(mode == HALF_DUPLEX)
	{
		HDstateControl(in);	// half-duplex state control
	}
}

////////////////////////////////////////////////////
//
// Hardware device configuration setting functions
//
void Audio::setMic()
{
	main_dlg->EnableAudioCheck();
	main_dlg->SetAudioCheck(TRUE);

	switch(state)
	{
	case stateMic:
		break;

	case stateSpeaker:
		main_dlg->SetSpeakerInd(0);
		bva->close();
		ncsAppSendData(ncsApp, -1, STOP, NULL, 0);
		audioReadInit();
		audioReadStart();
		break;

	case stateBoth:
		state = stateMic;
		bva->close();
		ncsAppSendData(ncsApp, -1, STOP, NULL, 0);
		break;
	}

	state = stateMic;
}

void Audio::setSpeaker()
{
	main_dlg->EnableAudioCheck();
	main_dlg->SetAudioCheck(FALSE);

	switch(state)
	{
	case stateMic:
		audioReadClose();
		audioReadStop();
		main_dlg->SetMicInd(0);
		bva->init();
		break;

	case stateSpeaker:
		break;

	case stateBoth:
		audioReadClose();
		audioReadStop();
		break;
	}

	state = stateSpeaker;
}

void Audio::setBoth()
{
	main_dlg->EnableAudioCheck(FALSE);
	main_dlg->SetAudioCheck(TRUE);

	switch(state)
	{
	case stateMic:
		bva->init();
		break;

	case stateSpeaker:
		audioReadInit();
		audioReadStart();
		break;

	case stateBoth:
		break;
	}

	state = stateBoth;
}

////////////////////////////////////////////////
//
// Full-duplex state control
//
void Audio::FDstateControl(StateInput in)
{
	if(automatic)
	{
		if(state == stateMic)
		{
			if(in == Abut)
			{
				automatic = 0;
				setBoth();
			}
			
			if((in == sig) || (in == Sbut))
			{
				setSpeaker();
			}
		}
		else
		if(state == stateSpeaker)
		{
			if(in == Abut)
			{
				automatic = 0;
				setBoth();
			}
			
			if((in == Nsig) || (in == Sbut))
			{
				setMic();
			}
		}

		if(in == Mbut)
		{
			automatic = 0;
			manual = 1;
		}

	}
	else
	{
		if(in == Abut)
		{
			automatic = 1;
			setMic();
		}

		if(in == Mbut)
		{
			if(manual == 0)
			{
				manual = 1;
				setSpeaker();
			}
			else
			{
				manual = 0;
				setBoth();
			}
		}

		if(manual)
		{
			if((state == stateSpeaker) && (in == Sbut))
				setMic();
			else
			if((state == stateMic) && (in == Sbut))
				setSpeaker();
		}
	}
}

////////////////////////////////////////////////
//
// Half-duplex state control
//
void Audio::HDstateControl(StateInput in)
{
	if(automatic)
	{
		if(state == stateMic)
		{
			if(in == Abut)
			{
				automatic = 0;
			}
			
			if((in == sig) || (in == Sbut))
			{
				setSpeaker();
			}
		}
		else
		if(state == stateSpeaker)
		{
			if(in == Abut)
			{
				automatic = 0;
			}
			
			if((in == Nsig) || (in == Sbut))
			{
				setMic();
			}
		}

		if(in == Mbut)
		{
			automatic = 0;
			manual = 1;
		}
	}
	else
	{
		if(state == stateMic)
		{
			if(in == Abut)
			{
				automatic = 1;
				manual = 0;
			}
			
			if(in == Sbut)
			{
				setSpeaker();
			}
		}
		else
		if(state == stateSpeaker)
		{
			if(in == Abut)
			{
				automatic = 1;
				manual = 0;
				setMic();
			}
			
			if(in == Sbut)
			{
				setMic();
			}
		}

		if(in == Mbut)
		{
			manual = (int)!(BOOL)manual;
		}
	}
}



