/*********************************************************************
 *
 * File:         control_clbks.c
 *
 * Description:  Messages callbacks
 *
 * Author:       Tomasz Stachowiak, NPAC
 *
 * Date:         10-14-96
 *
 *********************************************************************/
#include "stdafx.h"

#include <stdio.h>

#include "global.h"
#include "session_manage.h"
#include "BuenaVista.h"
#include "BuenaVistaDlg.h"
#include "Conference.h"
#include "daemon.h"


extern int tmpSocket;
extern int mbrSocket[];
extern int inviteSocket;

extern int dbState;
PROCESS_INFORMATION rec_info;
PROCESS_INFORMATION sim_info;

Participant inviting;
int sendInvitation = 0;
int inviteHandle=12345678;

FILE *log;

/**************************************************************************
 *
 * Callbacks for local control messages
 *
 **************************************************************************/

/* Session messages */

void ncsSessionCreate()
{
  DWORD time;
  int id;
  DWORD leng = 20;
  Participant initiator;

  initiator.handle = 0;
  if(!GetUserName(initiator.user, &leng))
	  strcpy(initiator.user, "unknown");
  initiator.host = gethostid();
  
  time = GetTickCount();
  id = time % 256;
  setSessionID(id);

  sessionCreate(&initiator);

  sessionSetLocalMember(0);
}

void ncsSessionAddMember(Participant *part)
{
  InviteData *data;
  int no;
  int size;

  PRINTF(log, "Engine: invitation send to %s@%d\n", part->user, part->host);
  no = sessionGetNoMembers();
  if(no > 0)
    size = (no - 1) * sizeof(Participant) + sizeof(InviteData);
  else
    size = sizeof(InviteData);
  data = (InviteData *)malloc(size);
  data->handle = htonl(getLocalHandle());
  data->numMembers = no;
  sessionGetMembers(data->parts);

  if(!sendNewMessage(part, reqNewRemoteSessionInvite, 
			   (char *)data, size))
    {
      PRINTF(log, "Engine: cannot connect\n");
      sendLocalMessage(indLocalSessionTimeout, (char *)part, sizeof(Participant));
    }
  else
    {
      PRINTF(log, "Engine: invitation received\n");
      sendLocalMessage(indLocalSessionConnected, (char *)part, sizeof(Participant));
    }
    

  free(data);
}

void ncsSessionJoin()
{
  sendInviteMessage(reqRemoteSessionAccept, NULL, 0);  

  mbrSocket[inviteHandle] = inviteSocket;
  inviteSocket = -1;
}

void ncsSessionReject()
{
  sessionInit();
  sendInviteMessage(reqRemoteSessionReject, NULL, 0);  
  closesocket(inviteSocket);
  inviteSocket = -1;
}

/*
void localClbSessionGetMembers(MsgHeader *header, char *msg, int len)
{
  InviteData *data;
  int no, size;

  no = sessionGetNoMembers();
  if(no > 0)
    size = (no - 1) * sizeof(Participant) + sizeof(InviteData);
  else
    size = sizeof(InviteData);
  data = (InviteData *)malloc(size);
  data->numMembers = no;
  sessionGetMembers(data->parts);
  sendLocalMessage(indLocalSessionMembers, 
			   (char *)data, size);
  free(data);
}
*/

void ncsSessionQuit()
{
  sendMemberMessage(-1, reqRemoteSessionQuit, NULL, 0);
  //sendAppMessage(-1, indLocalAppQuit, NULL, 0);
  //quitDaemon();
}

void ncsSendToApp(int no, int tag, char *msg, int len)
{
	AppData *data;
	int size;

	size = len + sizeof(AppData) - 1;
	data = (AppData *)new char[size];
	if(len > 0)
		memcpy(data->data, msg, len);
	data->tag = htons(tag);
	data->port = htonl(no);

	sendMemberMessage(-1, reqRemoteSessionSendToApp, (char *)data, size);

	delete data;
}

/*
void localClbSessionSendToApp(MsgHeader *header, char *msg, int len)
{
  AppData *data;
  int no;

  data = (AppData *)msg;
  no = getAppNumber(data->port);
  PRINTF(log, log, "App no %d, port %d\n", no, data->port);
  if(no >= 0)
    sendAppMessage(no, data->tag, data->data, len);
}

void localClbSessionStartApp(MsgHeader *header, char *msg, int len)
{
  int *num = (int *)msg;

  *num = htonl(*num);
  sendMemberMessage(-1, reqRemoteSessionStartApp, (char *)num, len);
}

void localClbSessionSaveInit(MsgHeader *header, char *msg, int len)
{
  char *ncspath;
  char path[100];

  if(dbState == DbStateIdle)
    {
      ncspath = getenv("NCS_PATH");
      strcpy(path, ncspath);
      strcat(path, "\\NCSrecord\\NCSrecord.exe");
      
      dbState = DbStateSave;
      start_process(path, &rec_info);

      sendAppMessage(-1, indDbStart, NULL, 0);
    }
}

void localClbSessionSaveClose(MsgHeader *header, char *msg, int len)
{
  if(dbState == DbStateSave)
    {
	  TerminateProcess(rec_info.hProcess, 0);
      dbState = DbStateIdle;
      
      sendAppMessage(-1, indDbStop, NULL, 0);
    }
}


void localClbSessionRestoreInit(MsgHeader *header, char *msg, int len)
{
  char *ncspath;
  char path[100];

  if(dbState == DbStateIdle)
    {
      ncspath = getenv("NCS_PATH");
      strcpy(path, ncspath);
      strcat(path, "\\NCSretrieve\\NCSretrieve.exe");
      
      dbState = DbStateRestore;
      start_process(path, &sim_info);
    }
}

void localClbSessionRestoreClose(MsgHeader *header, char *msg, int len)
{
  if(dbState == DbStateRestore)
    {
	  TerminateProcess(sim_info.hProcess, 0);
      dbState = DbStateIdle;
    }
}
*/

void ncsSessionReset()
{
	int i;
	MsgHeader hdr;

	sendMemberMessage(-1, reqRemoteSessionQuit, NULL, 0);
	for(i=0; i<MAX_MEMBERS; i++)
	{
		if(getReceiverData(i, &hdr))
		{
			sessionDeleteMember(i);
			closesocket(mbrSocket[i]);
			mbrSocket[i] = -1;
		}
    }
	//sendAppMessage(-1, indLocalAppQuit, NULL, 0);
	sessionInit();
}

/**************************************************************************
 *
 * Callbacks for new control messages
 *
 **************************************************************************/
void ncsSessionInit()
{

  sessionInit();
  /*
  localSocket = tmpSocket;
  tmpSocket = -1;

  
  if(sendInvitation)
    {
      sendLocalMessage(indLocalSessionInvited, 
			 (char *)&inviting, sizeof(Participant));  
      sendInvitation = 0;
    }
	*/
}

void newClbSessionConnect(struct sockaddr_in *addr, MsgHeader *header, char *msg, int len)
{
  mbrSocket[header->sender.handle] = tmpSocket;
  tmpSocket = -1;
  PRINTF(log, "Member %d connected, socket %d\n", header->sender.handle, mbrSocket[header->sender.handle]);
}

/*
void newClbAppAttach(struct sockaddr_in *addr, MsgHeader *header, char *msg, int len)
{
  int no;

  no = getFreeAppSocket();
  appSocket[no] = tmpSocket;
  tmpSocket = -1;
  PRINTF(log, log, "Engine: application port %d, number %d attached\n", *(unsigned short *)msg, no);
  sessionAddApp(no, *(unsigned short *)msg);
  if(dbState == DbStateSave)
    {
      sendAppMessage(no, indDbStart, NULL, 0);
    }
}
*/

void newClbSessionInvite(struct sockaddr_in *addr, MsgHeader *header, char *msg, int len)
{
  InviteData *data;

  data = (InviteData *)msg;
  inviteSocket = tmpSocket;
  inviteHandle = ntohl(data->handle);
  PRINTF(log, "Handle %ld\n", data->handle);
  PRINTF(log, "Handle %ld\n", inviteHandle);
  tmpSocket = -1;
  if(sessionCreated())
    {
      PRINTF(log, "Engine: session exists -- reject\n");
      sendInviteMessage(reqRemoteSessionReject, 
			     NULL, 0);
      closesocket(inviteSocket);
      inviteSocket = -1;

      return;
    }

  if(sessionInitialized())
    {
      PRINTF(log, "Engine: transmit invitation from %s@%d to manager\n", header->sender.user, header->sender.host);
      sendLocalMessage(indLocalSessionInvited, 
			 (char *)&header->sender, sizeof(Participant));  

      return;
    }

  if(sessionAbandoned())
    {
      PRINTF(log, "Engine: start manager\n");
      memcpy(&inviting, &header->sender, sizeof(Participant));
      //inviting.host = ntohl(inviting.host);
	  sendInvitation = 1;
      //startCM();

      return;
    }    
}


/**************************************************************************
 *
 * Callbacks for members control messages
 *
 **************************************************************************/
/**/
void remoteClbSessionAddRest(int no, MsgHeader *header, char *msg, int len)
{
  InviteData *data;
  int i, number, id;

  setSessionID(header->session);

  data = (InviteData *)msg;
  number = ntohl(data->numMembers);
  //data->parts[0].host = ntohl(data->parts[0].host);
  sessionCreate(&data->parts[0]);
  sessionSetLocalMember(data->parts[number-1].handle);

  if(no != 0)
    mbrSocket[0] = connect2member(data->parts[0].host);

  for(i=1; i<number; i++)
    {
     //data->parts[i].host = ntohl(data->parts[i].host);
     if(sessionAddMember(&data->parts[i]))
	{
	  id = data->parts[i].handle;
	  if((id != no) && (i < number - 1))
	    {
	      if(mbrSocket[id] >= 0)
		mbrSocket[getFreeMbrSocket()] = mbrSocket[id];
	      mbrSocket[id] = connect2member(data->parts[i].host);
	    }
	}
    }
}

/**/
void remoteClbSessionAccept(int no, MsgHeader *header, char *msg, int len)
{
  InviteData *data;
  int data_size, number;

  
  header->sender.handle = no;
  sendMemberMessage(-1, reqRemoteSessionAddMember, 
		     (char *)&header->sender, sizeof(Participant));

  sessionAddMember(&header->sender);

  number = sessionGetNoMembers();
  PRINTF(log, "Engine: Send add rest with %d members\n", number);
  if(number > 0)
    data_size = (number - 1) * sizeof(Participant) + sizeof(InviteData);
  else
    data_size = sizeof(InviteData);

  data = (InviteData *)malloc(data_size);
  data->numMembers = htonl(number);
  sessionGetMembers(data->parts);

  sendLocalMessage(indLocalSessionAccept, 
		     (char *)&header->sender, sizeof(Participant));

  sendMemberMessage(no, reqRemoteSessionAddRest,
			   (char *)data, data_size);
  free(data);
}

/**/
void remoteClbSessionReject(int no, MsgHeader *header, char *msg, int len)
{
  sendLocalMessage(indLocalSessionReject, 
		      (char *)&header->sender, sizeof(Participant));  
  closesocket(mbrSocket[no]);
  mbrSocket[no] = -1;
}

/**/
void remoteClbSessionAddMember(int no, MsgHeader *header, char *msg, int len)
{
	Participant *part = (Participant *)msg;
	//part->host = ntohl(part->host);
	sessionAddMember(part);
}

void remoteClbSessionQuit(int no, MsgHeader *header, char *msg, int len)
{
  sessionDeleteMember(no);
  closesocket(mbrSocket[no]);
  mbrSocket[no] = -1;
}  

/**/
void remoteClbSessionSendToApp(int no, MsgHeader *header, char *msg, int len)
{
  AppData *data;
  int size;

  data = (AppData *)msg;
  data->recID = ntohl(data->recID);
  data->port = ntohl(data->port);
  data->tag = ntohs(data->tag);
  size = len - sizeof(AppData) + 1;
  sendAppMessage(no, data->port, data->tag, data->data, size);
}

void remoteClbSessionSendAppData(int no, MsgHeader *header, char *msg, int len)
{
  AppData *data;

  data = (AppData *)msg;
  data->recID = ntohl(data->recID);
  data->port = ntohl(data->port);
  data->tag = ntohs(data->tag);
  sendAppMessage(no, data->port, indAppData, (char *)data, len);
}

/**/
void remoteClbSessionStartApp(int no, MsgHeader *header, char *msg, int len)
{
	int *num = (int *)msg;

	*num = ntohl(*num);
	sendLocalMessage(indLocalSessionStartApp, (char *)num, len);
}






